from common import *
import stim


def test_heralded_dem_simple():
    circuit_str = """\
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
MR 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
DEPOLARIZE1(0.02) 2
HERALDED_ERASE(0.01) 0
DETECTOR rec[-1]
M 0 2 4
DETECTOR(1, 1) rec[-2] rec[-3] rec[-6]
DETECTOR(3, 1) rec[-1] rec[-2] rec[-5]
OBSERVABLE_INCLUDE(0) rec[-1]\
"""
    circuit = stim.Circuit(circuit_str)
    dem = circuit.detector_error_model(approximate_disjoint_errors=True)
    print(dem)
    dem_str = """\
error(0.005) D2
error(0.005) D2 D3
error(0.01333333333333333) D3 D4
detector(1, 0) D0
detector(3, 0) D1
detector(1, 1) D3
detector(3, 1) D4
logical_observable L0\
"""
    assert dem.approx_equals(
        stim.DetectorErrorModel(dem_str),
        atol=mwpf.heralded_dem.DEM_MIN_PROBABILITY,
    )
    # the regular dem does not contain
    heralded_dem = mwpf.heralded_dem.HeraldedDetectorErrorModel.of(circuit)
    skeleton_dem = heralded_dem.skeleton_dem.dem()
    print("######### skeleton dem #########")
    print(skeleton_dem)
    assert skeleton_dem.approx_equals(
        stim.DetectorErrorModel(
            """\
error(6.661338147750937e-16) D3
error(0.01333333333333333) D3 D4
detector(1, 0) D0
detector(3, 0) D1
detector(1, 1) D3
detector(3, 1) D4
logical_observable L0\
"""
        ),
        atol=mwpf.heralded_dem.DEM_MIN_PROBABILITY,
    )
    # that skeleton edge must exist
    assert not skeleton_dem.approx_equals(
        stim.DetectorErrorModel(
            """\
detector(1, 0) D0
detector(3, 0) D1
detector(1, 1) D3
detector(3, 1) D4
logical_observable L0\
"""
        ),
        atol=mwpf.heralded_dem.DEM_MIN_PROBABILITY,
    )

    print("######### heralded DEM #########")
    print(heralded_dem)
    assert (
        str(heralded_dem)
        == """\
HeraldedDetectorErrorModel:
    skeleton hypergraph:
        D3: 6.661338e-16 ()
        D3, D4: 1.333333e-02 ()
    heralded hypergraph on D2:
        D3: 5.000000e-01 ()\
"""
    )


def test_heralded_dem_heralded_erase_example():
    circuit_str_with_comment = """\
HERALDED_ERASE(0.01) 0
# Declare a flag detector based on the erasure
DETECTOR rec[-1]

# Erase qubit 2 with 2% probability
# Separately, erase qubit 3 with 2% probability
HERALDED_ERASE(0.02) 2 3

# Do an XXXX measurement
MPP X2*X3*X5*X7
# Apply partially-heralded noise to the two qubits
HERALDED_ERASE(0.01) 2 3 5 7
DEPOLARIZE1(0.0001) 2 3 5 7
# Repeat the XXXX measurement
MPP X2*X3*X5*X7
# Declare a detector comparing the two XXXX measurements
DETECTOR rec[-1] rec[-6]
# Declare flag detectors based on the erasures
DETECTOR rec[-2]
DETECTOR rec[-3]
DETECTOR rec[-4]
DETECTOR rec[-5]\
"""
    circuit_str = """\
HERALDED_ERASE(0.01) 0
DETECTOR rec[-1]
HERALDED_ERASE(0.02) 2 3
MPP X2*X3*X5*X7
HERALDED_ERASE(0.01) 2 3 5 7
DEPOLARIZE1(0.0001) 2 3 5 7
MPP X2*X3*X5*X7
DETECTOR rec[-1] rec[-6]
DETECTOR rec[-2]
DETECTOR rec[-3]
DETECTOR rec[-4]
DETECTOR rec[-5]\
"""
    circuit = stim.Circuit(circuit_str_with_comment)
    print("######### original circuit #########")
    print(circuit)
    assert str(circuit) == circuit_str
    heralded_dem = mwpf.heralded_dem.HeraldedDetectorErrorModel.of(circuit)
    print("######### skeleton circuit #########")
    print(heralded_dem.skeleton_circuit)
    assert (
        str(heralded_dem.skeleton_circuit)
        == """\
DEPOLARIZE1(1e-15) 0
DEPOLARIZE1(1e-15) 2 3
MPP X2*X3*X5*X7
DEPOLARIZE1(1e-15) 2 3 5 7
DEPOLARIZE1(0.0001) 2 3 5 7
MPP X2*X3*X5*X7
DETECTOR abs[1] abs[0]\
"""
    )
    print("######### heralded DEM #########")
    print(heralded_dem)
    assert (
        str(heralded_dem)
        == """\
HeraldedDetectorErrorModel:
    skeleton hypergraph:
        D1: 2.666133e-04 ()
    heralded hypergraph on D2:
        D1: 5.000000e-01 ()
    heralded hypergraph on D3:
        D1: 5.000000e-01 ()
    heralded hypergraph on D4:
        D1: 5.000000e-01 ()
    heralded hypergraph on D5:
        D1: 5.000000e-01 ()\
"""
    )


def test_heralded_dem_heralded_pauli_channel_1_example():
    circuit_str_with_comment = """\
# With 10% probability perform a phase flip of qubit 0.
HERALDED_PAULI_CHANNEL_1(0, 0, 0, 0.1) 0
DETECTOR rec[-1]  # Include the herald in detectors available to the decoder

# With 20% probability perform a heralded dephasing of qubit 0.
HERALDED_PAULI_CHANNEL_1(0.1, 0, 0, 0.1) 0
DETECTOR rec[-1]

# Subject a Bell Pair to heralded noise.
MXX 0 1
MZZ 0 1
HERALDED_PAULI_CHANNEL_1(0.01, 0.02, 0.03, 0.04) 0 1
MXX 0 1
MZZ 0 1
DETECTOR rec[-1] rec[-5]  # Did ZZ stabilizer change?
DETECTOR rec[-2] rec[-6]  # Did XX stabilizer change?
DETECTOR rec[-3]    # Did the herald on qubit 1 fire?
DETECTOR rec[-4]    # Did the herald on qubit 0 fire?\
"""
    circuit_str = """\
HERALDED_PAULI_CHANNEL_1(0, 0, 0, 0.1) 0
DETECTOR rec[-1]
HERALDED_PAULI_CHANNEL_1(0.1, 0, 0, 0.1) 0
DETECTOR rec[-1]
MXX 0 1
MZZ 0 1
HERALDED_PAULI_CHANNEL_1(0.01, 0.02, 0.03, 0.04) 0 1
MXX 0 1
MZZ 0 1
DETECTOR rec[-1] rec[-5]
DETECTOR rec[-2] rec[-6]
DETECTOR rec[-3]
DETECTOR rec[-4]\
"""
    circuit = stim.Circuit(circuit_str_with_comment)
    print("######### original circuit #########")
    print(circuit)
    assert str(circuit) == circuit_str
    heralded_dem = mwpf.heralded_dem.HeraldedDetectorErrorModel.of(circuit)
    print("######### skeleton circuit #########")
    print(heralded_dem.skeleton_circuit)
    assert (
        str(heralded_dem.skeleton_circuit)
        == """\
PAULI_CHANNEL_1(0, 0, 1e-15) 0
PAULI_CHANNEL_1(0, 0, 1e-15) 0
MXX 0 1
MZZ 0 1
PAULI_CHANNEL_1(1e-15, 1e-15, 1e-15) 0 1
MXX 0 1
MZZ 0 1
DETECTOR abs[3] abs[1]
DETECTOR abs[2] abs[0]\
"""
    )
    print("######### heralded DEM #########")
    print(heralded_dem)
    assert (
        str(heralded_dem)
        == """\
HeraldedDetectorErrorModel:
    skeleton hypergraph:
        D2: 1.998401e-15 ()
        D2, D3: 1.998401e-15 ()
        D3: 1.998401e-15 ()
    heralded hypergraph on D4:
        D2: 2.000000e-01 ()
        D2, D3: 3.000000e-01 ()
        D3: 4.000000e-01 ()
    heralded hypergraph on D5:
        D2: 2.000000e-01 ()
        D2, D3: 3.000000e-01 ()
        D3: 4.000000e-01 ()\
"""
    )


def test_heralded_dem_add_detectors():
    circuit_str = """\
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
MR 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
HERALDED_ERASE(0.01) 0 2 1
M 0 2 4\
"""
    circuit = stim.Circuit(circuit_str)
    heralded_dem = mwpf.heralded_dem.HeraldedDetectorErrorModel.of(circuit)
    assert len(heralded_dem.heralded_measurements) == 3
    assert len(heralded_dem.detected_heralded_measurements) == 0  # none detected
    assert len(heralded_dem.undetected_heralded_measurements) == 3
    assert heralded_dem.heralded_detectors == (None, None)
    assert heralded_dem.heralded_detector_indices == tuple()
    assert heralded_dem.detector_id_to_herald_id == {}
    assert heralded_dem.num_heralds == 0
    # add heralded detectors
    circuit_with_herald_detected = mwpf.heralded_dem.add_herald_detectors(circuit)
    print(circuit_with_herald_detected)
    assert (
        str(circuit_with_herald_detected)
        == """\
R 0 1 2 3 4
TICK
CX 0 1 2 3
TICK
CX 2 1 4 3
TICK
MR 1 3
DETECTOR(1, 0) rec[-2]
DETECTOR(3, 0) rec[-1]
HERALDED_ERASE(0.01) 0 2 1
DETECTOR rec[-3]
DETECTOR rec[-2]
DETECTOR rec[-1]
M 0 2 4\
"""
    )
    heralded_dem = mwpf.heralded_dem.HeraldedDetectorErrorModel.of(
        circuit_with_herald_detected
    )
    assert len(heralded_dem.heralded_measurements) == 3
    assert len(heralded_dem.detected_heralded_measurements) == 3  # all detected
    assert len(heralded_dem.undetected_heralded_measurements) == 0
    assert len(heralded_dem.heralded_detectors) == 5
    assert heralded_dem.heralded_detectors[:2] == (None, None)
    assert heralded_dem.heralded_detectors[2] != None
    assert heralded_dem.heralded_detectors[3] != None
    assert heralded_dem.heralded_detectors[4] != None
    assert heralded_dem.heralded_detector_indices == (2, 3, 4)
    assert heralded_dem.detector_id_to_herald_id == {2: 0, 3: 1, 4: 2}
    assert heralded_dem.num_heralds == 3



from test_basic import *


def test_erasure_decoding():
    vertex_num = 3
    weighted_edges = [
        mwpf.HyperEdge([0, 1], 100),
        mwpf.HyperEdge([1, 2], 100),
        mwpf.HyperEdge([2, 0], 100),
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    solver.solve(mwpf.SyndromePattern([0, 2]))
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [2]
    assert bound.lower == bound.upper
    assert bound.upper == 100
    # however if I set the first two edges as erasures, it should choose the first two edges
    # because their weights becomes 0
    solver.solve(mwpf.SyndromePattern([0, 2], erasures=[0, 1]))
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper == 0


def test_override_weights_decoding():
    vertex_num = 3
    weighted_edges = [
        mwpf.HyperEdge([0, 1], 100),
        mwpf.HyperEdge([1, 2], 100),
        mwpf.HyperEdge([2, 0], 100),
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    # force set weights to [0, 0, 100]
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            override_weights=[0, 0, 100],
            override_ratio=1,
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper == 0
    # mix weights of ratio = 0.8
    # 0.8 * [0, 0, 100] + 0.2 * [100, 100, 100] = [20, 20, 100]
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            override_weights=[0, 0, 100],
            override_ratio=0.8,
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper.approx_eq(40)
    # set to negative weight
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            override_weights=[-20, 10, 100],
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper == -10
    # if I do not mix the weights that much, the optimal solution is still [2]
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            override_weights=[-20, 10, 100],
            override_ratio=0.1,
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [2]
    assert bound.lower == bound.upper
    assert bound.upper == 100
    # mixing the negative weights enough and then the optimal solution changes
    # 0.5 * [-20, 10, 100] + 0.5 * [100, 100, 100] = [40, 55, 100]
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            override_weights=[-20, 10, 100],
            override_ratio=0.5,
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper == 95


def test_heralded_decoding():
    vertex_num = 3
    weighted_edges = [
        mwpf.HyperEdge([0, 1], 1),
        mwpf.HyperEdge([1, 2], 1),
        mwpf.HyperEdge([2, 0], 1),
    ]
    heralds = [
        # add another set of errors with 50% error on edge 0 and 50% error on edge 1
        [(0, 0), (1, 0)],
        # some probability mix
        [(0, 0.2), (1, 0.3)],
        [(0, 0.6), (1, 0.5)],
        {0: -0.2, 1: -0.3},
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges, heralds=heralds)
    print(initializer)
    assert initializer.heralds == [
        {0: 0, 1: 0},
        {0: 0.2, 1: 0.3},
        {0: 0.6, 1: 0.5},
        {0: -0.2, 1: -0.3},
    ]
    # when the first herald error happens, it adds two 0-weighted edges
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            heralds=[0],
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper == 0
    # when the second herald error happens, it adds weight 0.2 and weight 0.3 edges
    # mixing weights 0.2 and 1 gives us log((1 + exp(0.2) * exp(1)) / (exp(0.2) + exp(1))) = 0.092181801390
    #     one can verify that by p1 = 1 / (1 + exp(0.2)) = 45%, p2 = 1 / (1 + exp(1)) = 27%
    #        p' = p1 (1-p2) + p2 (1-p1) = 47.7%, w' = log((1-p')/p') ~= 0.092
    # mixing weights 0.3 and 1 gives us log((1 + exp(0.3) * exp(1)) / (exp(0.3) + exp(1))) = 0.13782240494
    # two edges together weight 0.23000420633778762
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            heralds=[1],
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper == 0.23000420633778762
    # try some negative edges
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    solver.solve(
        mwpf.SyndromePattern(
            [0, 2],
            heralds=[3],
        )
    )
    subgraph, bound = solver.subgraph_range()
    print(subgraph)
    print((bound.lower, bound.upper))
    assert subgraph == [0, 1]
    assert bound.lower == bound.upper
    assert bound.upper == -0.23000420633778762


def test_exclusive_weight_sum():
    assert mwpf.exclusive_weight_sum(0.2, 1) == 0.09218180139025334
    assert mwpf.exclusive_weight_sum(0.3, 1) == 0.13782240494753428
    assert mwpf.exclusive_weight_sum(0.3, 0) == 0
    assert mwpf.exclusive_weight_sum(0, 0.3) == 0
    assert mwpf.exclusive_weight_sum(-1, 0.3) == -0.13782240494753428
    assert mwpf.exclusive_weight_sum(-1, 0) == 0
    assert mwpf.exclusive_weight_sum(-0.2, 1) == -0.09218180139025334
    assert mwpf.exclusive_weight_sum(-0.3, 1) == -0.13782240494753428



import os
import json
import string
import random
from typing import Optional
from IPython.display import display, Javascript, HTML

print(__file__)
FOLDER = os.path.dirname(os.path.abspath(__file__))
JS_PATH = os.path.join(FOLDER, "dist", "hyperion-visual.compressed.js")
assert (os.path.exists(JS_PATH), "please run `node run build` first")
with open(JS_PATH, "r", encoding="utf-8") as file:
    library_js = file.read()

library_loaded = False


def require_library() -> None:
    global library_loaded
    if library_loaded:
        return
    display(HTML("""<script type="module" id='hyperion_visual_compressed_js_caller'>
                           /* HYPERION_VISUAL_MODULE_CODE_BEGIN */
                           """ + library_js + """
                       /* HYPERION_VISUAL_MODULE_CODE_END */
                         </script>
                       """
                 ))
    library_loaded = True


def add_random_div(width: Optional[str] = None) -> str:
    id = ''.join(random.choices(string.ascii_lowercase, k=16))
    style = ""
    if width is not None:
        style = f'style="width: {width};"'
    display(HTML(f'<div {style} id="{id}"></div>'))
    return id


default_matrix = {
    "edges": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14],
    "hs": 0,
    "table": [
        ["", "0", "1", "2", "3", "4", "5", "6", "7", "8",
         "9", "1\n0", "1\n1", "1\n2", "1\n3", "1\n4", "="],
        ["0.", "1", "", "", "1", "", "", "", "",
         "1", "", "", "", "1", "", "", "1"],
        ["1.", "", "", "", "", "", "", "1", "1",
         "", "", "", "", "", "", "", "1"],
        ["2.", "", "1", "1", "", "", "", "",
         "", "", "", "", "", "", "", "", ""],
        ["3.", "", "", "1", "1", "1", "", "",
         "", "", "", "", "", "", "", "", ""],
        ["4.", "", "", "", "", "1", "1", "1",
         "", "", "", "", "", "", "", "", ""],
        ["5.", "1", "1", "", "", "", "", "", "",
         "", "", "", "", "", "", "1", ""],
        ["6.", "", "", "", "", "", "1", "", "",
         "1", "1", "", "", "", "", "", ""],
        ["7.", "", "", "", "", "", "", "", "1",
         "", "1", "", "", "", "", "", ""],
        ["8.", "", "", "", "", "", "", "", "",
         "", "", "", "", "", "1", "1", ""],
        ["9.", "", "", "", "", "", "", "", "",
         "", "", "", "1", "1", "1", "", ""],
        ["10.", "", "", "", "", "", "", "", "",
                "", "", "1", "1", "", "", "", ""]
    ],
    "version": "0.0.1",
    "is_echelon_form": False
}

js_call_main = '''
function sleep(ms) {
  return new Promise(resolve => setTimeout(resolve, ms));
}
async function wait_library() {
  for (let i = 0; i < 10; ++i) {
    if (globalThis.hyperion_visual != null) {
      break
    }
    if (i != 0) console.log(`window.hyperion_visual not ready, tried ${i} times`)
    await sleep(100)
  }
  if (globalThis.hyperion_visual != null) {
    main()
  } else {
    throw new Error("window.hyperion_visual, failed after 1s from window.onload")
  }
}
wait_library()
'''


def display_test_matrix(matrix=None) -> None:
    require_library()
    id = add_random_div()
    parity_matrix_data = default_matrix
    if matrix is not None:
        parity_matrix_data = matrix.to_visualize_json()
    code = "let display_matrix = " + json.dumps(parity_matrix_data) + '''
function main() {
    globalThis.hyperion_visual.parity_matrix.bind_to_div("#''' + id + '''", display_matrix)
}
    ''' + js_call_main
    # print(code)
    display(Javascript(code))


def display_test_hypergraph(width: Optional[str] = "50%") -> None:
    require_library()
    id = add_random_div(width=width)
    code = '''
function main() {
    globalThis.hyperion_visual.hypergraph.bind_to_div("#''' + id + '''")
}
    ''' + js_call_main
    # print(code)
    display(Javascript(code))



import sys
import os

static_patches = [
    (
        "Cargo.toml",
        [
            ('name = "mwpf"', 'name = "mwpf_$name$"', 3),
            ('default-run = "mwpf"', 'default-run = "mwpf_$name$"', 1),
        ],
    ),
    (
        "src/lib.rs",
        [
            ("fn mwpf(", "fn mwpf_$name$(", 1),
        ],
    ),
    (
        "src/main.rs",
        [
            ("use mwpf::cli::*;", "use mwpf_$name$::cli::*;", 1),
        ],
    ),
    (
        "src/python/mwpf/sinter_decoders.py",
        [
            ("import mwpf\n", "import mwpf_$name$\n", 1),
            ("from mwpf import", "from mwpf_$name$ import", 1),
            ("getattr(mwpf, decoder_type)", "getattr(mwpf_$name$, decoder_type)", 1),
            ("SinterMWPFDecoder", "SinterMWPF$Name$Decoder", 4),
            ("SinterHUFDecoder", "SinterHUF$Name$Decoder", 1),
            ("SinterSingleHairDecoder", "SinterSingleHair$Name$Decoder", 1),
            ("MwpfCompiledDecoder", "Mwpf$Name$CompiledDecoder", 3),
        ],
    ),
    (
        "src/python/mwpf/ref_circuit.py",
        [
            ("import mwpf\n", "import mwpf_$name$\n", 1),
            ("mwpf.", "mwpf_$name$.", None),
        ],
    ),
    (
        "src/python/mwpf/heralded_dem.py",
        [
            ("import mwpf\n", "import mwpf_$name$\n", 1),
            ("mwpf.", "mwpf_$name$.", None),
        ],
    ),
    (
        "src/python/mwpf/__init__.py",
        [
            ("from .mwpf import *", "from .mwpf_$name$ import *", 1),
            ("mwpf.", "mwpf_$name$.", 2),
            ("mwpf,", "mwpf_$name$,", 1),
        ],
    ),
    (
        "tests/python/test_sinter.py",
        [
            ("SinterMWPFDecoder", "SinterMWPF$Name$Decoder", None),
            ("SinterHUFDecoder", "SinterHUF$Name$Decoder", None),
        ],
    ),
    (
        "README.md",
        [
            ("pip install -U mwpf\n", "pip install -U mwpf_$name$\n", 1),
            (
                "pip install -U 'mwpf[stim]'\n",
                "pip install -U 'mwpf_$name$[stim]'\n",
                1,
            ),
            ('decoders = ["mwpf"],', 'decoders = ["mwpf_$name$"],', 1),
            (
                '"mwpf": SinterMWPFDecoder',
                '"mwpf_$name$": SinterMWPF$Name$Decoder',
                2,
            ),
            ("import SinterMWPFDecoder", "import SinterMWPF$Name$Decoder", 1),
            ("from mwpf import ", "from mwpf_$name$ import ", 2),
        ],
    ),
]

####### module name patches #######
pyclass_patch_files = [
    "src/dual_module.rs",
    "src/example_codes.rs",
    "src/html_export.rs",
    "src/mwpf_solver.rs",
    "src/util_py.rs",
    "src/util.rs",
    "src/visualize.rs",
    "src/matrix/interface.rs",
    "src/matrix/row.rs",
]
for filename in pyclass_patch_files:
    static_patches.append(
        (
            filename,
            [('pyclass(module = "mwpf"', 'pyclass(module = "mwpf_$name$"', None)],
        ),
    )


def patches_of(name: str) -> list:
    def patch(new: str):
        return (
            new.replace("$name$", name.lower())
            .replace("$Name$", name.capitalize())
            .replace("$NAME$", name.upper())
        )

    patches = [
        (name, [(old, patch(new), count) for old, new, count in lst])
        for name, lst in static_patches
    ]
    if name == "rational":
        patches.append(
            (
                "pyproject.toml",
                [
                    ('name = "mwpf"', 'name = "mwpf_rational"', 1),
                    ("f64_weight", "rational_weight", 2),
                ],
            )
        )
    elif name == "incr":
        patches.append(
            (
                "pyproject.toml",
                [
                    ('name = "mwpf"', 'name = "mwpf_incr"', 1),
                    (
                        "python_binding,f64_weight,embed_visualizer",
                        "python_binding,f64_weight,incr_lp,embed_visualizer",
                        1,
                    ),
                    ('"f64_weight",', '"f64_weight" ,"incr_lp",', 1),
                ],
            )
        )
    elif name == "fast":
        patches.append(
            (
                "pyproject.toml",
                [
                    ('name = "mwpf"', 'name = "mwpf_fast"', 1),
                    (
                        "python_binding,f64_weight,embed_visualizer",
                        "python_binding,f64_weight,fast_ds,embed_visualizer",
                        1,
                    ),
                    ('"f64_weight",', '"f64_weight" ,"fast_ds",', 1),
                ],
            )
        )
    return patches


# patch is strict
def patch(dry: bool, name: str):
    for filename, replacements in patches_of(name):
        with open(filename, "r") as f:
            content = f.read()
        # check occurrences first
        for old, new, occurrence in replacements:
            assert (
                occurrence is None or content.count(old) == occurrence
            ), f"count {filename} for '{old}': {content.count(old)} != {occurrence}"
            assert (
                content.count(new) == 0
            ), f"count {filename} for '{new}': {content.count(new)} != 0"
        # during application of the replacements, also check occurrence
        for old, new, occurrence in replacements:
            assert occurrence is None or content.count(old) == occurrence
            assert content.count(new) == 0
            old_content = content
            content = content.replace(old, new)
            assert (
                content != old_content
            ), f"Patch failed for {filename}: {old} -> {new}"
        # check occurrences last
        for old, new, occurrence in replacements:
            assert occurrence is None or content.count(new) == occurrence
            assert content.count(old) == 0
        if not dry:
            with open(filename, "w") as f:
                f.write(content)
    if not dry:
        # up to here, all files has been checked and updated, rename the src/python/mwpf folder
        os.rename("src/python/mwpf", f"src/python/mwpf_{name}")


# revert is best-practice
def revert(name: str):
    # first change the folder back
    os.rename(f"src/python/mwpf_{name}", "src/python/mwpf")
    for filename, replacements in patches_of(name):
        with open(filename, "r") as f:
            content = f.read()
        for old, new, occurrence in replacements:
            count = content.count(new)
            if occurrence is not None and count != occurrence:
                print(
                    f"[warning] reverting process counting error '{old}' '{new}' {occurrence} != {count}"
                )
            content = content.replace(new, old)
        with open(filename, "w") as f:
            f.write(content)


if __name__ == "__main__":
    assert (
        len(sys.argv) == 3
    ), "Usage: python pyproject-patch.py [rational|incr|fast] [dry|apply|revert]"
    assert sys.argv[1] in ["rational", "incr", "fast"]
    if sys.argv[2] == "dry":
        patch(dry=True, name=sys.argv[1])
    elif sys.argv[2] == "apply":
        patch(dry=True, name=sys.argv[1])
        patch(dry=False, name=sys.argv[1])
    elif sys.argv[2] == "revert":
        revert(name=sys.argv[1])
    else:
        raise ValueError("Invalid argument, should be dry|apply|revert")



from common import *
import fusion_blossom as fb
import numpy as np


def prepare_hyperion_solver() -> mwpf.SolverSerialJointSingleHair:
    vertex_num = 6
    weighted_edges = [
        mwpf.HyperEdge([0, 1], 100),
        mwpf.HyperEdge([1, 2], 100),
        mwpf.HyperEdge([2, 3], 100),
        mwpf.HyperEdge([3, 4], 100),
        mwpf.HyperEdge([4, 5], 100),
        mwpf.HyperEdge([1, 2, 3], 60),  # hyper edge
        mwpf.HyperEdge([0], 0),  # virtual vertex
        mwpf.HyperEdge([5], 0),  # virtual vertex
    ]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.SolverSerialJointSingleHair(initializer)
    return solver


def prepare_fusion_solver() -> fb.SolverSerial:
    vertex_num = 6
    weighted_edges = [(0, 1, 100), (1, 2, 100), (2, 3, 100), (3, 4, 100), (4, 5, 100)]
    virtual_vertices = [0, 5]
    initializer = fb.SolverInitializer(vertex_num, weighted_edges, virtual_vertices)
    solver = fb.SolverSerial(initializer)
    return solver


def test_compare_hyperion_fusion() -> None:
    syndrome = [1, 2, 4]
    # hyperion
    hyperion = prepare_hyperion_solver()
    hyperion.solve(mwpf.SyndromePattern(syndrome))
    hyperion_subgraph = hyperion.subgraph()
    _, bound = hyperion.subgraph_range()
    print(hyperion_subgraph)
    print((bound.lower, bound.upper))
    # fusion blossom
    fusion = prepare_fusion_solver()
    fusion.solve(fb.SyndromePattern(syndrome))
    fusion_subgraph = fusion.subgraph()
    print(fusion_subgraph)


def test_initializer_uniform_weight():
    code = mwpf.CodeCapacityColorCode(d=5, p=0.01)
    initializer = code.get_initializer()
    initializer.uniform_weights(weight=1)
    for edge in initializer.weighted_edges:
        assert edge.weight.float() == 1


def test_numpy_array_parse():
    # pytest -s tests/python/test_basic.py::test_numpy_array_parse
    syndrome = np.array([1, 2, 4], dtype="int32")
    mwpf.SyndromePattern(syndrome)


def test_non_connected_graphs():
    vertex_num = 6
    weighted_edges = [mwpf.HyperEdge([1, 2, 3], 60)]
    initializer = mwpf.SolverInitializer(vertex_num, weighted_edges)
    solver = mwpf.Solver(initializer)
    print(solver.get_initializer())
    solver.solve(mwpf.SyndromePattern([1, 2, 3]))
    subgraph = solver.subgraph()
    print(subgraph)
    assert subgraph == [0]
    solver.clear()
    solver.solve(mwpf.SyndromePattern([]))
    subgraph = solver.subgraph()
    print(subgraph)
    assert subgraph == []



from .mwpf_fast import *


__doc__ = mwpf_fast.__doc__
if hasattr(mwpf_fast, "__all__"):
    __all__ = mwpf_fast.__all__  # type: ignore

try:
    from .sinter_decoders import *
    from . import heralded_dem
    from . import ref_circuit
except BaseException as e:
    # raise e
    ...



from common import *
import stim
import sinter
import numpy as np


params = dict(panic_action=mwpf.PanicAction.RAISE)


@pytest.mark.parametrize(
    "decoder",
    [
        mwpf.SinterMWPFFastDecoder(cluster_node_limit=50, **params),
        mwpf.SinterHUFFastDecoder(**params),
        mwpf.SinterMWPFFastDecoder(cluster_node_limit=0, **params),
    ],
)
@pytest.mark.parametrize(
    "p",
    [0.001],
)
@pytest.mark.parametrize(
    "d",
    [3],
)
@pytest.mark.parametrize(
    "code_type",
    [
        "color_code:memory_xyz",
        "repetition_code:memory",
        "surface_code:rotated_memory_x",
    ],
)
def test_sinter_decode(decoder: sinter.Decoder, p: float, d: int, code_type: str):
    if "mwpf" not in sys.modules:
        print("[skip] because sinter would fail to import without the mwpf package")
        print(
            "once https://github.com/quantumlib/Stim/pull/873 is merged, this test should be enabled"
        )
        return

    # pytest -s tests/python/test_sinter.py::test_sinter_decode
    circuit = stim.Circuit.generated(
        code_type,
        rounds=d,
        distance=d,
        after_clifford_depolarization=p,
        before_round_data_depolarization=p,
        before_measure_flip_probability=p,
        after_reset_flip_probability=p,
    )

    task = sinter.Task(
        circuit=circuit,
        collection_options=sinter.CollectionOptions(max_shots=10, max_errors=10),
    )

    sinter.collect(
        num_workers=2,
        tasks=[task],
        decoders=["mwpf"],
        custom_decoders={"mwpf": decoder},
    )


def test_sinter_heralded_error():
    circuit = stim.Circuit(
        """\
R 0 1
X_ERROR(0.01) 0  # with lower probability, an error happens at 0
X_ERROR(0.1) 1  # normally we would always guess that the actual error happens at 1
HERALDED_PAULI_CHANNEL_1(0.02, 0.02, 0, 0) 0  # D0
DETECTOR rec[-1]
MPP Z0*Z1  # D1
DETECTOR rec[-1]
M 0
OBSERVABLE_INCLUDE(0) rec[-1]  # read qubit 0
"""
    )
    print(circuit)
    dem = circuit.detector_error_model(approximate_disjoint_errors=True)
    print("######### dem #########")
    print(dem)
    # without special handling, the dem includes the heralded error as regular pauli error
    assert dem == stim.DetectorErrorModel(
        """
error(0.02) D0
error(0.02) D0 D1 L0
error(0.1) D1
error(0.01) D1 L0
"""
    )
    decoder = mwpf.SinterMWPFFastDecoder(**params)
    compiled_decoder = decoder.compile_decoder_for_dem(dem=dem)
    # trigger both D0 and D1:
    bit_packed_detection_event_data = np.packbits(
        np.array([[1, 1]]), axis=-1, bitorder="little"
    )
    prediction = compiled_decoder.decode_shots_bit_packed(
        bit_packed_detection_event_data=bit_packed_detection_event_data
    )
    observables = compiled_decoder.predictor.get_observable_bits(prediction)
    assert observables == [1]

    decoder.with_circuit(circuit)
    compiled_decoder = decoder.compile_decoder_for_dem(dem=dem)
    prediction = compiled_decoder.decode_shots_bit_packed(
        bit_packed_detection_event_data=bit_packed_detection_event_data
    )
    observables = np.unpackbits(
        prediction[0], count=dem.num_observables, bitorder="little"
    )
    assert observables == [1]

    """
    In the above example, it seems like heralded errors are handled correctly in both cases.
    Let's see if we have construct cases where the regular fail but the normal one works.
    
    It's probably easier to not trigger the heralded detection. Normal handling of the error might
    treat the non-triggered event as "even parity check", and thus selecting two highly improbable
    cases
    """

    circuit2 = stim.Circuit(
        """\
R 0 1 2 3 4 5 6 7 8  # data qubits
MPP X0*X3*X6  # measure initial logical state
MPP X0*X1*X3*X4 X1*X2 X4*X5*X7*X8 X6*X7 # prepare stabilizer state
#  0  1  2
#  3  4  5
#  6  7  8

# add some noise
HERALDED_ERASE(0.3) 4  # the center qubits
DETECTOR rec[-1]
Z_ERROR(0.0001) 3 5  # these errors are less likely than the center erasure,
                   # however, if the herald indicator is False, we should still use this

# when the detector is not triggered, a regular decoder might think a Y error happens
MPP X0*X1*X3*X4 X1*X2 X4*X5*X7*X8 X6*X7
DETECTOR rec[-4] rec[-9]
DETECTOR rec[-3] rec[-8]
DETECTOR rec[-2] rec[-7]
DETECTOR rec[-1] rec[-6]
MPP Z0*Z3 Z1*Z2*Z4*Z5 Z3*Z4*Z6*Z7 Z5*Z8
DETECTOR rec[-4]
DETECTOR rec[-3]
DETECTOR rec[-2]
DETECTOR rec[-1]
MPP X0*X3*X6
OBSERVABLE_INCLUDE(0) rec[-1] rec[-15]
"""
    )
    print(circuit2)
    dem = circuit2.detector_error_model(approximate_disjoint_errors=True)
    print("######### dem #########")
    print(dem)
    assert dem == stim.DetectorErrorModel(
        """
error(0.075) D0
error(0.075) D0 D1 D3
error(0.075) D0 D1 D3 D6 D7
error(0.075) D0 D6 D7
error(0.0001) D1 L0
error(0.0001) D3
detector D2
detector D4
detector D5
detector D8
"""
    )
    decoder = mwpf.SinterMWPFFastDecoder(**params)
    compiled_decoder = decoder.compile_decoder_for_dem(dem=dem)
    # trigger D1 and D3
    bit_packed_detection_event_data = np.packbits(
        np.array([[0, 1, 0, 1, 0, 0, 0, 0, 0]]), axis=-1, bitorder="little"
    )
    prediction = compiled_decoder.decode_shots_bit_packed(
        bit_packed_detection_event_data=bit_packed_detection_event_data
    )
    observables = compiled_decoder.predictor.get_observable_bits(prediction)
    assert observables == [0]
    # on the other hand, a heralded-error-aware decoder should be able to figure out
    # that although the middle error is has higher probability, when the herald detector is False,
    # it doesn't really have any effect and thus we should choose the two error(0.0001) despite
    # that the probability of their both happening is pretty low
    decoder.with_circuit(circuit2)
    compiled_decoder = decoder.compile_decoder_for_dem(dem=dem)
    prediction = compiled_decoder.decode_shots_bit_packed(
        bit_packed_detection_event_data=bit_packed_detection_event_data
    )
    observables = np.unpackbits(
        prediction[0], count=dem.num_observables, bitorder="little"
    )
    assert observables == [1]
